#define STR_NAME "Yubico"
#define STR_VERSION "0.1"
#define STR_AUTHOR "Michael Bonfils"


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <syslog.h>

#include <sys/types.h>
#include <pwd.h>
#include <grp.h>


#include <linux/input.h>

#include <ykclient.h>

typedef struct {
	char *conffile;

	int client_id;
	char *api_key;
	char *url;

	char *user;
	char *group;
	char *on_sucess;
	char *on_fail;
	char *on_quit;
} conf;

typedef struct {
	int from_code;
	int to_char;
} item;

// Maybe there is something very very more simpler than that ? maybe libhid ?
item array[] = {
	{ KEY_1, '1' },
	{ KEY_2, '2' },
	{ KEY_3, '3' },
	{ KEY_4, '4' },
	{ KEY_5, '5' },
	{ KEY_6, '6' },
	{ KEY_7, '7' },
	{ KEY_8, '8' },
	{ KEY_9, '9' },
	{ KEY_0, '0' },
	{ KEY_A, 'a' },
	{ KEY_B, 'b' },
	{ KEY_C, 'c' },
	{ KEY_D, 'd' },
	{ KEY_E, 'e' },
	{ KEY_F, 'f' },
	{ KEY_G, 'g' },
	{ KEY_H, 'h' },
	{ KEY_I, 'i' },
	{ KEY_J, 'j' },
	{ KEY_K, 'k' },
	{ KEY_L, 'l' },
	{ KEY_M, 'm' },
	{ KEY_N, 'n' },
	{ KEY_O, 'o' },
	{ KEY_P, 'p' },
	{ KEY_Q, 'q' },
	{ KEY_R, 'r' },
	{ KEY_S, 's' },
	{ KEY_T, 't' },
	{ KEY_U, 'u' },
	{ KEY_V, 'v' },
	{ KEY_W, 'w' },
	{ KEY_X, 'x' },
	{ KEY_Y, 'y' },
	{ KEY_Z, 'z' },
	{ KEY_TAB, '\t' },
	{ KEY_ENTER, '\n' },
	{ -1, -1 } };



void freesecure( char **X ) {
	if ( *X ) free( *X );
	*X = NULL;
}

char convert_code2char( int code ) {
	int i=0;
	while ( array[i].from_code != -1 && array[i].from_code != code )
		i++;
	if ( array[i].to_char == -1 )
		syslog( LOG_ERR, "Failed to convert code %d\n", code );
	return array[i].to_char;
}


conf* init_conf() {
	conf *c = (conf*) malloc( sizeof(conf) );
	memset( c, 0, sizeof(conf) );

	c->conffile = strdup( "/etc/yubico.conf" );

	return c;
}

int read_conf( conf *current_conf, int argc, char *argv[]  ) {
	int opt;
	// at first, try to find if user has overriden conffile
	while ((opt = getopt(argc, argv, "c:")) != -1) {
		switch( opt ) {
			case 'c':
				freesecure(&current_conf->conffile);
				current_conf->conffile = strdup(optarg);
				break;
			case 'v':
				fprintf( stdout, "%s - version %s - %s\n", STR_NAME, STR_VERSION, STR_AUTHOR );
				return -1;
				break;
		}
	}
	

	FILE *fp = fopen( current_conf->conffile, "r" );
	if ( fp == NULL ) {
		syslog( LOG_ERR, "Failed to open %s, aborting", current_conf->conffile );
		return( -1 );
	}
	
	char line[128], field[128], value[128];
	while ( fgets(line, sizeof(line), fp) ) {
		if ( strlen(line)==0 || line[0]=='#' ) continue;
		
		*field = *value = '\0';
		int r = sscanf( line, "%128[A-Z_-]=%128s", field, value );
		if ( r != 2 ) continue;

		syslog( LOG_ERR, "'%s'='%s'", field, value );

		if ( strcmp( field, "CLIENT_ID" ) == 0 ) 
			current_conf->client_id = atoi( value );
		else if ( strcmp( field, "API_KEY" ) == 0 )
			current_conf->api_key = strdup( value );
		else if ( strcmp( field, "RUN_ON_SUCCESS" ) == 0 ) 
			current_conf->on_sucess = strdup( value );
		else if ( strcmp( field, "RUN_ON_FAILURE" ) == 0 )
			current_conf->on_fail = strdup( value );
		else if ( strcmp( field, "RUN_ON_QUIT" ) == 0 )
			current_conf->on_quit = strdup( value );
		else if ( strcmp( field, "USER" ) == 0 )
			current_conf->user = strdup( value );
		else if ( strcmp( field, "GROUP" ) == 0 )
			current_conf->group = strdup( value );
	}
	fclose( fp );
	
	// check if client_id and api_key are set
	if ( current_conf->client_id == 0 || current_conf->api_key == NULL || strlen(current_conf->api_key) == 0 ) {
		syslog( LOG_ERR, "Missing CLIENT_ID and API_KEY in %s", current_conf->conffile );
		return -1;
	}

	return 0;
}


int set_uid_gid( const char *user, const char *group ) {
	int r;
	if ( user == NULL || strlen( user ) == 0 )
		return -1;

	struct passwd *uid = getpwnam( user );
	if ( !uid ) {
		syslog( LOG_ERR, "Failed to extract user info for '%s'", user );
		return -1;
	}

	r = setuid( uid->pw_uid );
	if ( r != 0 ) {
		syslog( LOG_ERR, "Failed to change UID %d : %d %s", uid->pw_uid, errno, strerror( errno ) );
	}

	int gid = uid->pw_gid;
	if ( group != NULL && strlen( group ) > 0 )  {
		struct group *grp = getgrnam( group );
		if ( grp == NULL ) {
			syslog( LOG_ERR, "Failed to extract group info for '%s'", group );
		} else {
			gid = grp->gr_gid;
		}
	}
	
	r = setgid( gid );
	if ( r != 0 ) {
		syslog( LOG_ERR, "Failed to change GID %d : %d %s", gid, errno, strerror( errno ) );
	}

	return 0;
}

void launch_script( const char *user, const char *group, const char* script) {
	int rc = fork();
	if ( rc == 0 ) { // child
		daemon( 1, 0 );
		setsid();
		// change username
		if ( set_uid_gid( user, group ) != 0 ) {
			syslog( LOG_ERR, "Failed to use %s:%s", user, group );
			exit( 0 );
		}

		execve( script, NULL, NULL );
		syslog( LOG_ERR, "execve has failed : %d %s", errno, strerror( errno ) );
		exit( 0 );
	} else if ( rc == -1 ) {
		syslog( LOG_ERR, "Failed to fork : %d %s\n", errno, strerror( errno ) );
	}
}

int check_OTP( ykclient_t *yk, conf *current_conf, const char *otp ) {
	syslog( LOG_ERR, "OTP=>%s\n", otp );
	int rt = ykclient_request( yk, otp );
	if ( rt != YKCLIENT_OK ) {
		syslog( LOG_ERR, "ykclient_request has failed : %d %s\n", rt, ykclient_strerror( rt ) );
		launch_script( current_conf->user, current_conf->group, current_conf->on_fail );
		return -1;
	}

	syslog( LOG_ERR, "OTP is success\n" );
	launch_script( current_conf->user, current_conf->group, current_conf->on_sucess );
	return 0;
}

int main( int argc, char *argv[] ) {
	openlog( "Yubico", LOG_PID, LOG_USER );

	conf *current_conf = init_conf();

	if ( read_conf( current_conf, argc, argv ) != 0 )
		return -1;

#ifdef DEBUG
	extern char **environ;
	char **ptr = environ;
	while ( *ptr ) {
		syslog( LOG_ERR, "%s", *ptr );
		ptr++;
	}
#endif

	// TODO, parse all entries in /dev/input to found correct item for YubiKey if DEVNAME is empty
	const char *input = getenv("DEVNAME"); // DEVNAME is something like /dev/input/event2
	if ( input == NULL ) {
		syslog( LOG_ERR, "Yubico : No DEVNAME provided, exiting..." );
		exit( 0 );
	}

	syslog( LOG_ERR, "Yubico : %s", input );

	ykclient_t *yk = NULL;

	int rt = ykclient_init( &yk );
	if ( rt != YKCLIENT_OK ) {
		syslog( LOG_ERR, "ykclient_init has failed : %d %s\n", rt, ykclient_strerror( rt ) );
		exit( -1 );
	}
	
	rt = ykclient_set_client_b64( yk, current_conf->client_id, current_conf->api_key );
	if ( rt != YKCLIENT_OK ) {
		syslog( LOG_ERR, "ykclient_set_client_b64 has failed : %d %s\n", rt, ykclient_strerror( rt ) );
		exit( -1 );
	}

	int fd = -1, retry = 0;

 	// trying to open input, may be not available the first time
	while ( fd == -1 ) {
		fd= open( input, O_RDONLY );
		if ( fd == -1 ) {
			if ( retry>5 ) {
				syslog( LOG_ERR, "Failed to open %s : %d %s\n", input, errno, strerror( errno ) );
				exit( -1 );
			}
			retry++;
			sleep( 1 );
		}
	}

	// try to grab device, excluding the input of YubiKey to be on other input
	// check /usr/include/linux/input.h
	int rc = ioctl( fd, EVIOCGRAB, 1 );
	if ( rc != 0 ) {
		syslog( LOG_ERR, "Failed to grab %s, event will be shared : %d %s\n", input, errno, strerror( errno ) );
	}


	// check Documentation/input/input.txt
	struct input_event mem;

	char current_otp[ 128 ];
	int  current_idx = 0;

	time_t last_event = 0;

	int quit = 0;
	while ( !quit ) {
		int r = read( fd, &mem, sizeof(mem) );
		if ( r>0 && r != sizeof(mem) ) {
			syslog( LOG_ERR, "partial read\n" );
		} else if ( r == sizeof(mem) ) {

			if ( mem.type == EV_KEY && mem.value == 1 ) {
				int c = convert_code2char( mem.code );
				
				if ( c>0 ) {
					if ( mem.time.tv_sec - last_event > 1 ) 
						// more than one second between the last event, reset the current OTP
						// FIXME: check if possible to use somes more accurates timings (like 50ms or something like that)
						current_idx=0;
					if ( current_idx + 1 < sizeof(current_otp) ) {
						last_event = mem.time.tv_sec;
						current_otp[ current_idx++ ] = c;
						if ( current_idx == 44 ) {
							current_otp[ current_idx ] = '\0';

							check_OTP( yk, current_conf, current_otp );
						}
					}
				}
			}


		} else if ( r == 0 ) {
			syslog( LOG_ERR, "empty read\n" );
		} else {
			syslog( LOG_ERR, "error has occured : %d %s\n", errno, strerror( errno ) );
			quit = 1;
		}
	}

	launch_script( current_conf->user, current_conf->group, current_conf->on_quit );
	ykclient_done( &yk );
	close( fd );
	return 0;
}
